
import java.util.HashMap;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.regex.*;
import java.io.IOException;

@SuppressWarnings("all") public class PBOSupport {
	
	//parameters
	public final int height;
	public final int width;
	//choices
	// choices with ids
	
	private static HashMap<String,String> pboArguments;
	private static final Pattern nameValue = Pattern.compile("--(.+)=(.+)");
	private static String[] commandLineArguments = null;
	private static String callMainMethodClass = null;
	
	private static PBOSupport pbo_support = null;
	
	public long startTime;
	
	private static void setCommandLineArguments(String[] args) {
		commandLineArguments = args;
		
		// do a pass through the command line arguments
		// so that we can call the project's main method
		// without having to create an instance of this
		// singleton first
		
		pbo_support = new PBOSupport();
	}
	
	public static PBOSupport getInstance() {
		if (pbo_support == null) {
			throw new IllegalStateException("The PBOSupport class has not been intialized.  The PBOSupport main method should be the point of entry for project.");
		}
		
		return pbo_support;
	}
	
	private PBOSupport() {
		pboArguments = new HashMap<String,String>();
		
		// we cannot initialize these variables within a method
		// so we must do them in the constructor
		
		// params with defaults


		// params with ranges


		// params with categories


		// choices with ids

		
		
		// parse the command line and optional options file BEFORE
		// we validate that all the pbo constructs have been initialized
		
		try { parseCommandLine(commandLineArguments); }
		catch (Exception e) {
			System.out.println("PBOSupport: something went wrong with reading the options file");
			e.printStackTrace();
			System.exit(-1);
		}
		
		pbo_check_param_values();
		
		// we will be able to tell, at weaving time, what these
		// types of variables are, so there is no concern here to
		// record these types in some data structure.
		//
		// furthermore, we test out parse* BEFORE we try and 
		// assign it to the variable so that we can warn the 
		// user of malformed input

		// parameters
		try {  Integer.parseInt(pboArguments.get("height")); } catch (NumberFormatException e) { System.out.println("Supplied value for height must be a long integer"); System.exit(-1); }
		height = Integer.parseInt(pboArguments.get("height"));
		try {  Integer.parseInt(pboArguments.get("width")); } catch (NumberFormatException e) { System.out.println("Supplied value for width must be a long integer"); System.exit(-1); }
		width = Integer.parseInt(pboArguments.get("width"));
		callMainMethodClass = pboArguments.get("javaClass");
		
		pbo_check_param_ranges();
		pbo_check_param_categories();
		pbo_check_choice_categories();
		
	}
	
	private void pbo_check_param_values() {
		// we assume that parseCommandLine has been called already
		
		// we can know at weaving time which params/choices do
		// not have defaults, so we force the user to provide
		// values for them first
		String[] defaultlessParams  = {"height","width"};
		String[] choices			= {};
		String mainMethodClassString= "javaClass";
		for(int i=0;i<defaultlessParams.length; i++) {
			String param = defaultlessParams[i];
			
			if (!pboArguments.containsKey(param)) {
				System.out.format("ERROR: Parameter --%s must be set.%n",param);
				System.exit(-1);
			}
		}
		
		for(int i=0; i<choices.length; i++) {
			String choice = choices[i];
			
			if(!pboArguments.containsKey(choice)) {
				System.out.format("ERROR: Choice --%s must be set%n",choice);
				System.exit(-1);
			}
		}
		
		if (!pboArguments.containsKey(mainMethodClassString)) {
			System.out.println("ERROR: Project entry point --javaClass must be set.");
			System.exit(-1);
		}
	}
	
	private void pbo_check_param_ranges() {
	}
	
	private void pbo_check_param_categories() {
	}
	
	private void pbo_check_choice_categories() {
	}
	
	private static void parseCommandLine(String[] args) throws IOException {
		// holds all the command line instantiations
		HashMap<String,String> commandLineArguments = new HashMap<String,String>();
		HashMap<String,String> optionsFileArguments = new HashMap<String,String>();
		HashMap<String,HashMap<String,Integer>> choiceNamedInstances = new HashMap<String,HashMap<String,Integer>>(0);
		
		for(String arg: args) {
			if (arg.equals("--")) break;
			
			// we do not expect to see more than one of these
			if (arg.startsWith("@")) {
				String optFile = arg.substring(1); // chop @
				optionsFileArguments = parseOptionsFile(optFile);
			} else if (arg.startsWith("--")) {
				Matcher matcher = nameValue.matcher(arg);
				if(matcher.find()) {
					String name = matcher.group(1);
					String value = matcher.group(2);
					
					if (choiceNamedInstances.containsKey(name) && ((HashMap<String,Integer>)choiceNamedInstances.get(name)).containsKey(value)) {
						Integer val = ((HashMap<String,Integer>)choiceNamedInstances.get(name)).get(value);
						commandLineArguments.put(name,val.toString());
					} else {
						commandLineArguments.put(name,value);
					}
				}
			} else {
				System.out.format("WARNING: Unrecognized argument format: %s. Skipping.%n", arg);
			}
			
		}
		
		// because putAll overrides old values with new values for the
		// same key we simply abuse that fact.
		//
		// because we prioritize command line argument values over option
		// file values, we put the command line arguments last
		
		pboArguments.putAll(optionsFileArguments);
		pboArguments.putAll(commandLineArguments);
	}
	
	private static HashMap<String,String> parseOptionsFile(String pathToFile) throws IOException {
		HashMap<String,String> optionsFileArguments = new HashMap<String,String>();
		BufferedReader optionsFile = null;
		
		// do not greedily grab until we see '=' because choices with
		// ids have lines with two '='s in them.
		
		Pattern nameValue = Pattern.compile("\\s*--(.+?)\\s*=\\s*(.+)");
		
		try {
			optionsFile = new BufferedReader(new FileReader(pathToFile));
			String line = null;
			
			while((line = optionsFile.readLine()) != null) {
				// no use doing an expensive regex if we don't have to
				if (!line.startsWith(";")) {
					Matcher matcher = nameValue.matcher(line);
					if(matcher.find()) {
						String name = matcher.group(1);
						String value = matcher.group(2);
						
						// this is a really bad hack to get around the fact that
						// lines about choices with ids have a ; comment at the end
						// 
						// and java's regex doesn't support conditionals within the
						// regular expression.
						//
						// ex.
						// --processing=4 ; = alternative
						//
						// ideal the regular expression is
						// s*--(.+?)=(.+)s*(?(?=;)(?:.+))
						//
						// which means optional start of white space
						// --(name)=(value)
						// followed by optional white space
						// if followed by ; then ignore till end of line
						// 
						// the draw back to this simple regular expression is that if
						// value turns out to be a string or character that contains
						// ; in it, this regular expression will truncate everything after
						// the semicolon within the string.  i'm hoping the chances 
						//of that are small, and we can fix it if it ever arises.
						
						if (value.lastIndexOf(";") != -1) {
							value = value.substring(0, value.lastIndexOf(";")).trim();
						}
						
						// when parsing the opt file, characters like " and ' stick
						// around even though they do NOT when parsing the command line
						//
						// therefore, we need to remove these beginning and end characters
						
						if (value.startsWith("\"") && value.endsWith("\"")) {
							value = value.substring(1, value.length()-1);
						}
						
						if (value.startsWith("'") && value.endsWith("'")) {
							value = value.substring(1,value.length()-1);
						}
						
						optionsFileArguments.put(name,value);
					}
				}			
			}
		} finally {
			optionsFile.close();
		}
		
		return optionsFileArguments;
	}
	
	private static String[] buildMainProgramArguments(String[] args) {
		String[] mainArguments = null;
		boolean shouldCopy = false;
		int copyIndex=0;
		
		// unlike C, args does not contain the name of the program
		// as its first argument
		for(int i=0; i<args.length; i++) {
			// arguments to the main program are only supposed to 
			// come after "--"
			if (args[i].equals("--")) {
				shouldCopy=true;
				mainArguments = new String[args.length-i-1];
				
				// we do NOT want to fall through to the if
				// statement below since we do NOT want to
				// copy over "--" as well
				
				continue;
			}
			
			if (shouldCopy) { mainArguments[copyIndex++] = args[i]; }
		}	
		
		// if we don't find "--", then no main program arguments were
		// supplied, but Java specs requires us to at least give it
		// an empty array
		
		if (mainArguments == null) { return new String[0]; }
		
		return mainArguments;
	}
	
	// choices should be represented by booleans
	// here, we default to false unless one of "true", "1", "active" or "on"
	// is the keyword supplied
	// therefore, we do not throw a CouldNotParse error.
	public static boolean parseBoolean(String bool) {
		if (bool.equals("true")||bool.equals("1")||bool.equals("active")||bool.equals("on")) {
			return true;
		}
		
		return false;
	}
	
	// main point of entry for a PbO enriched source
	public static void main(String[] args) throws Throwable {
		setCommandLineArguments(args);
		
		String[] passThroughArguments = buildMainProgramArguments(args);
		
		getInstance().startTime = System.currentTimeMillis();
		
		try {
			Class.forName(callMainMethodClass).getMethod("main",String[].class).invoke(null,(Object)passThroughArguments);
		} catch (Exception e) {
			if(e instanceof ClassNotFoundException) { System.out.format("PBOSupport: provided Class %s could not be found.%n",callMainMethodClass); }
			if(e instanceof NoSuchMethodException) { System.out.format("PBOSupport: Class %s has no main method to call.%n", callMainMethodClass); }
			if(e instanceof SecurityException) { System.out.format("PBOSupport: PBOSupport doe snot have public access to Class %s or to its main method.%n", callMainMethodClass); }
			if(e instanceof IllegalArgumentException) { System.out.format("PBOSupport: There was an error in parsing out the main program's arguments.  Please immediately contact the maintainers of PbO.%n"); }
			if(e instanceof IllegalAccessException) { System.out.format("PBOSupport: Could not access the main method of Class %s.%n",callMainMethodClass); }
			if(e instanceof java.lang.reflect.InvocationTargetException) { throw e.getCause(); }
			
			e.printStackTrace();
			System.exit(-1);
		}
	
	}
}
